/* vim: set ts=4 sw=4 noet ff=unix fileencoding=utf-8:
 *
 * Minimal multi-threaded path-tracer
 *
 * Copyright (c) 2011-2013 Andrej Bukosek.
 * All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <float.h>
#include <math.h>

#include <new>
#include <vector>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "profiler.h"

#define MAX3(a,b,c) ( ((a)>(b) && (a)>(c)) ? (a) : ((b)>(c) ? (b) : (c)) )

/*
 * 3D vector
 */
struct Vector {
    double x, y, z;

    inline Vector(double _x = 0.0, double _y = 0.0, double _z = 0.0) :
		x(_x), y(_y), z(_z) { }

	inline Vector operator+(const Vector &v) const {
		return Vector(x + v.x, y + v.y, z + v.z);
	}
	inline Vector operator-(const Vector &v) const {
		return Vector(x - v.x, y - v.y, z - v.z);
	}
	inline Vector operator*(double d) const {
		return Vector(x * d, y * d, z * d);
	}

	inline void operator+=(const Vector &v) { x += v.x; y += v.y; z += v.z; }
	inline void operator*=(double d)        { x *= d;   y *= d;   z *= d; }

	inline Vector operator-() const {
		return Vector(-x, -y, -z);
	}

	inline Vector norm() const {
		return *this * (1.0 / sqrt(x*x + y*y + z*z));
	}

	inline double dot(const Vector &v) const {
		return x*v.x + y*v.y + z*v.z;
	}

	inline Vector cross(const Vector &v) const {
		return Vector(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x);
	}
};

/*
 * 32-bit RGB color
 */
struct Color {
    float r, g, b;

	inline Color(float _r = 0.0f, float _g = 0.0f, float _b = 0.0f) :
		r(_r), g(_g), b(_b) { }

	inline Color operator+(const Color &c) const {
		return Color(r + c.r, g + c.g, b + c.b);
	}
	inline Color operator-(const Color &c) const {
		return Color(r - c.r, g - c.g, b - c.b);
	}
	inline Color operator*(const Color &c) const {
		return Color(r * c.r, g * c.g, b * c.b);
	}
	inline Color operator*(float f) const {
		return Color(r * f, g * f, b * f);
	}

	inline void operator+=(const Color &c) { r += c.r; g += c.g; b += c.b; }
	inline void operator*=(const Color &c) { r *= c.r; g *= c.g; b *= c.b; }
	inline void operator*=(float f)        { r *= f;   g *= f;   b *= f; }
};

/*
 * Marsaglia's xorshift RNG: http://www.jstatsoft.org/v08/i14/paper
 */
struct RNG {
    unsigned long long seed;

	inline RNG() : seed(88172645463325252ull) { }

	inline unsigned long long generate() {
		seed ^= seed << 13;
		seed ^= seed >>  7;
		seed ^= seed << 17;

		return seed;
	}

	inline double next() {
		return (double)generate() * (1.0  / (double)ULLONG_MAX);
	}

	inline float nextf() {
		return (float)generate() * (1.0f / (float)ULLONG_MAX);
	}
};

/*
 * 3D ray
 */
struct Ray {
    Vector origin, dir;

    Ray(const Vector &_o, const Vector &_d) : origin(_o), dir(_d) { }
};

typedef Ray Camera; /* :P */

/*
 * Supported materials
 */
enum Material {
	DIFFUSE,	/* Perfectly diffuse */
	SPECULAR,	/* Perfectly specular */
	REFRACTIVE	/* Dielectric */
};

/*
 * Sphere
 */
struct Sphere {
    double   r2;	/* Store squared radius for speed */
    Vector   pos;
    Color    c_emissive, c_diffuse;
    Material mat;
	double   ri_air, ri_obj;	/* Refractive indices */

	Sphere(double _r, const Vector &_pos,
		   const Color &_c_emi, const Color &_c_dif,
		   Material _mat, double ra = 1.0, double ro = 1.8) :
		r2(_r * _r), pos(_pos), c_emissive(_c_emi), c_diffuse(_c_dif),
		mat(_mat), ri_air(ra), ri_obj(ro) {  }

	/*
	 * Does the ray intersect this sphere?
	 * Returns distance if it does, 0.0 otherwise.
	 */
    double intersect(const Ray &r) const
    {
    	/* Solve the ray-sphere intersection equation */
        const Vector p_o = pos - r.origin;
		const double b   = p_o.dot(r.dir);
		double       det = b*b - p_o.dot(p_o) + r2;

		if (det < 0.0)
			return 0.0;		/* No solution -- ray doesn't intersect sphere */

		det = sqrt(det);

		double t = b - det;

		if (t > DBL_EPSILON)
			return t;

		t = b + det;

		if (t > DBL_EPSILON)
			return t;

		return 0.0;
    }
};

/* Store all spheres in the scene in this STL vector */
static std::vector<Sphere> objs;

/*
 * Check if given ray intersects any object in the scene.
 * Returns distance to object (t) and the object itself (obj).
 * The __restrict attributes help the compiler optimize the code better.
 */
static inline bool intersect_scene(const Ray &__restrict     r,
								   double *__restrict        t,
								   const Sphere **__restrict obj)
{
    size_t i = objs.size();

	*t = DBL_MAX;

    while (i--) {
        const double d = objs[i].intersect(r);

        if (d != 0.0 && d < *t) {
        	/* Object is nearer than our current best, store this one instead */
            *t   = d;
            *obj = &objs[i];
        }
    }

    return *t < DBL_MAX;
}

/* Intersects ray with scene and returns radiance at the point of intersection */
static Color get_radiance(const Ray &__restrict r_, RNG &__restrict rng)
{
    double        t;  /* Distance from ray to nearest intersection point */
	const Sphere *obj   = NULL;		/* Hit object */
    int           depth = 0;		/* Current 'recursion' depth */
    Ray           r     = r_;		/* Current ray */

    Color         lum(0.0f, 0.0f, 0.0f);	/* Luminance accumulated so far */
    Color         ref(1.0f, 1.0f, 1.0f);	/* Reflectance accumulated so far */

    for (;;) {
		/*
		 * Intersect ray with scene geometry.
		 * Return accumulated luminance if the ray missed.
		 */
        if (!intersect_scene(r, &t, &obj))
            return lum;

        /* Add object's emissive contribution */
		lum += ref * obj->c_emissive;

		/* Calculate reflectance contribution */
		Color ref_contrib = obj->c_diffuse;

		if (++depth > 4) {
			/* Choose maximum possible reflectivity */
			const float p = MAX3(ref_contrib.r, ref_contrib.g, ref_contrib.b);

			/* Perform Russian roulette */
			if (rng.nextf() < p && depth < 32)
				ref_contrib *= 1.0f / p;
			else
				return lum;
		}

		/* Blend accumulated reflectance and obj-contributed reflectance */
		ref *= ref_contrib;

		/* Calculate point of intersection */
		const Vector poi = r.origin + r.dir * t;

		/* Calculate intersection normal */
		const Vector n   = (poi - obj->pos).norm();

		/* Make sure the normal points away from the current ray's direction */
		const Vector nl  = n.dot(r.dir) < 0.0 ? n : -n;

        if (obj->mat == DIFFUSE) {
			/*
			 * Ideal diffuse reflection
			 */

            /* Choose a random direction in the hemisphere defined by the normal */
			const double r1  = 2.0 * M_PI * rng.next();
			const double r2  = rng.next();
			const double r2s = sqrt(r2);

			const Vector w = nl;
			Vector u;

			if (fabs(w.x) > DBL_EPSILON)
				u = Vector(0.0, 1.0, 0.0);
			else
				u = Vector(1.0, 0.0, 0.0);

			const Vector v = w.cross(u);

			const Vector d = (	u * cos(r1) * r2s +
								v * sin(r1) * r2s +
								w * sqrt(1.0 - r2)	).norm();

			/* Shoot a new ray in the chosen direction */
            r = Ray(poi, d);
        } else if (obj->mat == SPECULAR) {
			/*
			 * Ideal specular reflection
			 */

            r = Ray(poi, r.dir - n * 2.0 * n.dot(r.dir));
        } else {
			/*
			 * Ideal dielectric refraction
			 */

			/* Is the ray going into our object? */
            const bool into = n.dot(nl) > 0.0;

			/* Choose the correct refractive index */
            const double ra = obj->ri_air, ro = obj->ri_obj;
			const double rr = into ? ra/ro : ro/ra;

			/* Check for total internal reflection */
			const double d_n   = r.dir.dot(nl);
            const double cos2t = 1.0 - rr * rr * (1.0 - d_n * d_n);
            if (cos2t < 0.0) {
				/* Total internal reflection, return reflected ray */
                r = Ray(poi, r.dir - n * 2.0 * n.dot(r.dir));
                continue;
			}

            /* Calculate refracted direction */
            const Vector tdir = (r.dir * rr -
								 n * ( (into ? 1.0 : -1.0) *
									  (d_n * rr + sqrt(cos2t)) ) ).norm();

			/* Use Russian roulette to decide whether to reflect or refract */
			const double a  = ro - ra;
			const double b  = ro + ra;
			const double c  = 1.0 - (into ? -d_n : tdir.dot(n));
			const double R0 = (a * a) / (b * b);
			const double Re = R0 + (1.0 - R0)*c*c*c*c*c;
			const double p  = 0.25 + 0.5 * Re;

			if (rng.next() < p) {
				/* Reflection */
				ref *= (float)(Re / p);

				r = Ray(poi, r.dir - n * 2.0 * n.dot(r.dir));
			} else {
				/* Refraction */
				ref *= (float)((1.0 - Re) / (1.0 - p));

				r = Ray(poi, tdir);
			}
        }
    }
}


/* Recursively generate a 3D "sphereflake" fractal structure of given depth */
static void genSphereFlake(double x, double y, double z, double radius, int depth, int skip)
{
    if (depth == 0)
		return;

    // Sphere: radius, position, emission, color, material
    objs.push_back(Sphere(radius, Vector(x, y, z), Color(),
                          Color(fabsf(sinf(expf(cosf((float)x)))),
                                fabsf(sinf(expf(cosf((float)y)))),
                                fabsf(sinf(expf(cosf((float)z))))),
                          REFRACTIVE));

    const double rad2 = radius * 0.5;
    const double rad  = radius + rad2;

    if (skip != 1) genSphereFlake(x + rad, y, z, rad2, depth - 1, 2);
    if (skip != 2) genSphereFlake(x - rad, y, z, rad2, depth - 1, 1);
    if (skip != 3) genSphereFlake(x, y + rad, z, rad2, depth - 1, 4);
    if (skip != 4) genSphereFlake(x, y - rad, z, rad2, depth - 1, 3);
    if (skip != 5) genSphereFlake(x, y, z + rad, rad2, depth - 1, 6);
    if (skip != 6) genSphereFlake(x, y, z - rad, rad2, depth - 1, 5);
}

#define die(err) do { fprintf(stderr, "ERROR: %s\n", err); exit(1); } while (0)

int main(int argc, char *argv[])
{
    /*
     * Parse options
     */
    int   w = 1280, h = 720, sub = 2, super = 1, jobs = 0;
    float gamma = 2.2f;

    int ch;
    while ((ch = getopt(argc, argv, "d:g:j:s:S:")) != -1) switch (ch) {
        case 'd': /* Output image dimensions */
            if (sscanf(optarg, "%16dx%16d", &w, &h) != 2)
                die("Malformed dimensions (expected WxH).");

            if (w < 2 || h < 2)
                die("Dimensions out of range.");
            break;

        case 'g': /* Output image gamma factor */
            if (sscanf(optarg, "%32f", &gamma) != 1 || gamma < 1.0e-10f)
                die("Invalid argument for gamma correction factor.");
            break;

		case 'j': /* Number of render threads */
			if (sscanf(optarg, "%16d", &jobs) != 1 || jobs < 0)
                die("Invalid argument for number of threads.");
			break;

        case 's': /* Number of subsamples for anti-aliasing */
            if (sscanf(optarg, "%16d", &sub) != 1 || sub < 1)
                die("Invalid argument for number of subsamples.");
            break;

        case 'S': /* Number of supersamples for anti-aliasing */
            if (sscanf(optarg, "%16d", &super) != 1 || super < 1)
                die("Invalid argument for number of supersamples.");
            break;

        default:
            fprintf(stderr,
					"Usage: %s [-d WxH] [-g gamma] [-j threads] [-s sub] [-S super]\n",
					argv[0]);
            return 1;
    }

    /* Precalculate reciprocal values of often-used parameters */
	const double w_inv      = 1.0 / (double)w;
	const double h_inv      = 1.0 / (double)h;
	const double sub_inv    = 1.0 / (double)sub;


    /* Create framebuffer */
    Color *fb = new (std::nothrow) Color[w * h];

    if (!fb)
        die("Out of memory.");

    Profiler prof;

    fprintf(stderr,
			"Render: %dx%d at %d samples per pixel (sampling: %d super, %dx%d sub)\n",
            w, h, super*sub*sub, super, sub, sub);
#ifdef _OPENMP
    const int mp_ps = omp_get_num_procs();

	omp_set_num_threads(jobs == 0 ? mp_ps : jobs);

	const int mp_ts = omp_get_max_threads();

	fprintf(stderr, "OpenMP: %d core%s available, using %d thread%s\n",
			mp_ps, mp_ps > 1 ? "s" : "", mp_ts, mp_ts > 1 ? "s" : "");
#else
	fprintf(stderr, "OpenMP: disabled\n");
#endif


    /*
     * Generate scene
     */

	const Camera cam(Vector(50.0, 50.0, 250.0), Vector(-0.1, -0.1, -1.0).norm());
    const Vector cx = Vector((double)w / (double)h, 0.0, 0.0);
	const Vector cy = (cx.cross(cam.dir)).norm();

    // Light
    objs.push_back(Sphere(1000, Vector(0,-1100.0,0), Color(100,100,100), Color(), DIFFUSE));

    genSphereFlake(0, 0, 0, 50.0, 4, 0);

    fprintf(stderr, "Scene : %lu objects\n\n", objs.size());


    /*
     * Render
     */

    prof.start("render");
    /* Parallelize rendering into rows */
#pragma omp parallel for schedule(dynamic, 1)
    for (int y = 0; y < h; y++) {
#ifdef _OPENMP
		/* Output rendering progress only from first thread */
        if (omp_get_thread_num() == 0)
#endif
            fprintf(stderr, "\rRendering ... %3.2f%%", 100.0f * y / (h - 1));

        /* Seed the thread's RNG with a deterministic number (for debugging) */
		RNG rng;

		for (int x = 0; x < w; x++) {
			Color rad;	/* Accumulated radiance */
			float gw_sum = 0.0f;

			/* Perform subsampling */
			for (int sy = 0; sy < sub; sy++) {		/* Subpixel rows */
				for (int sx = 0; sx < sub; sx++) {	/* Subpixel cols */

					/* Perform supersampling */
					for (int s = 0; s < super; s++) {
						/* Generate jitter offsets */
						const double dx = ((double)sx + (-0.5 + rng.next())) * sub_inv;
						const double dy = ((double)sy + (-0.5 + rng.next())) * sub_inv;

						/* Calculate Gauss weight factor (sigma = 1) */
						const float gw = ((float)M_1_PI * 0.5f) *
						                 expf((float)(dx*dx + dy*dy) * -0.5f);

						gw_sum += gw;

						/* Calculate perturbed samples */
						const double sample_x = -0.5 + ((double)x + dx) * w_inv;
						const double sample_y = -0.5 + ((double)y + dy) * h_inv;

						/* Generate ray direction from perturbed samples */
						const Vector d = cam.dir + cx * sample_x + cy * sample_y;

						/* Shoot ray & accumulate radiance */
						rad += get_radiance(Ray(cam.origin, d.norm()), rng) * gw;
					}
				}
			}

			/* Account for super-/sub-sampling & copy to framebuffer */
			fb[(h - y - 1) * w + x] = rad * (1.0f / gw_sum);
		}
    }
    prof.stop("render");
	fprintf(stderr, "\rRendering ... 100.00%%\n");


    /*
     * Tone map the image (Ward)
     */

    const float Ldmax = 80.0f;		/* Max brightness of the display (sRGB) */
	const float G = 1.0f / gamma;	/* Gamma correction factor */

	const int   fpix = w * h;	/* Number of pixels in frame */

    prof.start("tonemap");
    /* Step 1: Calculate the logarithmic luminance average */
    float sum_log = 0.0f;
    for (int i = 0; i < fpix; i++) {
        const Color &p = fb[i];

		/* Calculate pixel's luminance */
		const float Y = logf(p.r * 0.2126f + p.g * 0.7152f + p.b * 0.0722f);

		if (finite(Y))
			sum_log += Y;
    }
    const float Llogavg = expf(sum_log / (float)fpix);

    /* Step 2: Calculate the scale factor */
    const float a = 1.219f + powf(Ldmax * 0.25f, 0.4f);
    const float b = 1.219f + powf(Llogavg, 0.4f);
    const float scale = powf(a / b, 2.5f) / Ldmax;

	/* Step 3: Process the frame */
#ifdef __INTEL_COMPILER
	/* ICC generates code which segfaults here, so disable vectorization */
#pragma novector
#endif
	for (int i = 0; i < fpix; i++) {
		/* Apply tone map scale factor & correct gamma */
		fb[i].r = powf(fb[i].r * scale, G);
		fb[i].g = powf(fb[i].g * scale, G);
		fb[i].b = powf(fb[i].b * scale, G);
	}
    prof.stop("tonemap");


    /* Finally, write a 16-bits-per-pixel PPM image */
    prof.start("file");
    FILE *fp = fopen("image.ppm", "w");
    fprintf(fp, "P6\n# bench_pt rendering\n");
    fprintf(fp, "# Render: %dx%d at %dspp (%d super, %dx%d sub)\n",
            w, h, super*sub*sub, super, sub, sub);
#ifdef _OPENMP
    fprintf(fp, "# OpenMP: enabled (%d threads)\n", mp_ts);
#else
    fprintf(fp, "# OpenMP: disabled\n");
#endif
    fprintf(fp, "# Scene : %lu objects\n", objs.size());
    fprintf(fp, "# Render time [s]: %.3f CPU / %.3f real\n",
            prof.Tcpu("render"), prof.Treal("render"));
    fprintf(fp, "# gamma   = %f\n# Llogavg = %f\n\n", gamma, Llogavg);
    fprintf(fp, "%d %d\n65535\n", w, h);

    for (int i = 0; i < fpix; i++) {
    	/* Clamp output values */
		unsigned int outR = (unsigned int)floorf(fb[i].r * 65536.0f);
		unsigned int outG = (unsigned int)floorf(fb[i].g * 65536.0f);
		unsigned int outB = (unsigned int)floorf(fb[i].b * 65536.0f);

		if (outR > 65535) outR = 65535;
		if (outG > 65535) outG = 65535;
		if (outB > 65535) outB = 65535;

		/* Write 16-bit values to file in big-endian order */
		fputc((outR & 0xff00) >> 8, fp);
		fputc( outR & 0x00ff,       fp);
		fputc((outG & 0xff00) >> 8, fp);
		fputc( outG & 0x00ff,       fp);
		fputc((outB & 0xff00) >> 8, fp);
		fputc( outB & 0x00ff,       fp);
    }

    fclose(fp);
    prof.stop("file");

    /* Print statistics */
    prof.print();

    delete[] fb;
    return 0;
}
