/*
 * Minimal profiling framework
 *
 * Copyright (c) 2012 Andrej Bukosek.
 * All rights reserved.
 */

#ifndef __PROFILER_H__
#define __PROFILER_H__

#include <stdio.h>
#include <string.h>

#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <list>

/*
 * Profiling timer abstraction
 */
struct Profiler {
private:
	struct Times {
		bool           inProgress;			/* Measurement in progress? */
		struct timeval startReal, endReal;	/* Wall clock (real) time */
		struct timeval startCPU, endCPU;	/* CPU time */
		clock_t        Treal, Tcpu;			/* Real time, CPU time [in us] */
	};

	typedef std::pair<char *, Times> MapPair;
	typedef std::list<MapPair> Map;

	Map timer;	/* Timer name -> data */


public:
	/* Start timer */
	void start(const char *name)
	{
		Times t;

		memset(&t, 0, sizeof(Times));
		t.inProgress = true;

		/* Save start times */
		gettimeofday(&t.startReal, NULL);

		struct rusage r;
		getrusage(RUSAGE_SELF, &r);
		t.startCPU = r.ru_utime;

		/* Create new timer entry, don't check for duplicates */
		timer.push_front(MapPair(strdup(name), t));
	}

	/* Stop timer */
	void stop(const char *name)
	{
		/* Check if the specified timer exists */
		Map::iterator it = timer.begin();

		for (; it != timer.end(); ++it)
			if (!strcmp(it->first, name))
				break;

		if (it != timer.end()) {
			/* It does */
			Times t = it->second;

			/* Save end times */
			struct rusage r;
			getrusage(RUSAGE_SELF, &r);
			t.endCPU = r.ru_utime;

			gettimeofday(&t.endReal, NULL);

			/* Calculate time difference */
			t.Treal = (clock_t)((t.endReal.tv_sec  - t.startReal.tv_sec)  * 1000 +
					  (t.endReal.tv_usec - t.startReal.tv_usec) / 1000);
			t.Tcpu  = (clock_t)((t.endCPU.tv_sec   - t.startCPU.tv_sec)   * 1000 +
					  (t.endCPU.tv_usec  - t.startCPU.tv_usec)  / 1000);

			t.inProgress = false;

			it->second = t;  /* Update entry */
		}
	}

	/* Retrieve times in seconds */
	float Treal(const char *name)
	{
		Map::const_iterator it = timer.begin();

		for (; it != timer.end(); ++it)
			if (!strcmp(it->first, name))
				return (float)it->second.Treal / 1000.0f;

		return -1.0f;
	}

	float Tcpu(const char *name)
	{
		Map::const_iterator it = timer.begin();

		for (; it != timer.end(); ++it)
			if (!strcmp(it->first, name))
				return (float)it->second.Tcpu / 1000.0f;

		return -1.0f;
	}

	/* Print a table with all finished timers */
	void print(FILE *fp = stderr, bool ansi = true) const
	{
		/* Print table header */
		fprintf(fp, "\n%sProfiling timers:%s\n",
				ansi ? "\033[1m" : "", ansi ? "\033[0m" : "");
		fprintf(fp, "\t%sName%27cT(CPU)[s]  T(real)[s] [hhh:mm:ss.ttt]%s\n",
				ansi ? "\033[4m" : "", ' ', ansi ? "\033[0m" : "");

		/* For each timer ... */
		for (Map::const_iterator i = timer.begin(); i != timer.end(); ++i) {
			const Times &t = i->second;

			/* Skip timers that are in progress */
			if (t.inProgress)
				continue;

			/* Convert the CPU & real time to seconds */
			const float Tc = (float)t.Tcpu  / 1000.0f;
			const float Tr = (float)t.Treal / 1000.0f;

			/* Construct a HHH:MM:SS.TTT format from real time */
			float r = Tr;
			const unsigned int h  = (unsigned int)(r / 3600.0f); r -= h * 3600;
			const unsigned int m  = (unsigned int)(r /   60.0f); r -= m * 60;
			const unsigned int s  = (unsigned int)(r);           r -= s;
			const unsigned int th = (unsigned int)(r * 1000.0f);

			/* Print table row: name, cpu time, real time, real time (nicer) */
			fprintf(fp, "\t%-30s%10.3f  %10.3f  %3u:%02u:%02u.%03u\n",
					i->first, Tc, Tr, h, m, s, th);
		}
	}
};

#endif /* __PROFILER_H__ */
